from __future__ import annotations

from types import SimpleNamespace
from unittest.mock import MagicMock

import pytest

@pytest.fixture()
def hook():
    from box_airflow_provider.hooks.box import BoxHook

    return BoxHook(box_conn_id="box_default")


@pytest.fixture()
def hook_with_fake(box_fake):
    from box_airflow_provider.hooks.box import BoxHook

    # noinspection PyTypeChecker
    return BoxHook(client_factory=box_fake.client)


def test_get_conn_creates_client_and_caches(hook):
    conn = MagicMock()
    conn.get_extra_dejson.return_value = {
        "client_id": "cid",
        "client_secret": "csecret",
        "enterprise_id": "eid",
    }

    hook.get_connection = MagicMock(return_value=conn)

    c1 = hook.get_conn()
    c2 = hook.get_conn()

    assert c1 is c2
    assert c1.auth.config.client_id == "cid"
    assert c1.auth.config.client_secret == "csecret"
    assert c1.auth.config.enterprise_id == "eid"
    hook.get_connection.assert_called_once_with("box_default")


def test_get_conn_raises_for_missing_credentials(hook):
    conn = MagicMock()
    conn.get_extra_dejson.return_value = {
        "client_id": "",
        "client_secret": "csecret",
        "enterprise_id": "eid",
    }
    hook.get_connection = MagicMock(return_value=conn)

    with pytest.raises(ValueError, match="Client ID, Client Secret and Enterprise ID must be provided"):
        hook.get_conn()


def test_test_connection_succeeds(hook_with_fake):

    ok, msg = hook_with_fake.test_connection()

    assert ok is True
    assert "successfully" in msg.lower()


def test_get_item_id_folder_and_file_resolution(hook_with_fake, box_fake):
    child_id = box_fake.seed_folder("/Parent/Child")
    file_id = box_fake.seed_file("/Parent/Child/file.txt")

    assert hook_with_fake.get_folder_id("/Parent/Child") == child_id
    assert hook_with_fake.get_file_id("/Parent/Child/file.txt") == file_id


def test_get_item_id_missing_folder_raises(hook_with_fake):
    with pytest.raises(FileNotFoundError, match="Folder 'Nope' not found"):
        hook_with_fake.get_folder_id("/Nope")


def test_get_existing_item_id_returns_deepest_existing(hook_with_fake, box_fake):
    a_id = box_fake.seed_folder("/A")

    item_id, existing_path = hook_with_fake.get_existing_item_id("/A/B/C", item_type="folder")
    assert item_id == a_id
    assert existing_path == "/A"


def test_get_item_id_root_folder_ok_and_root_file_error(hook_with_fake):
    assert hook_with_fake.get_folder_id("/") == "0"
    with pytest.raises(ValueError, match="Cannot get file ID for root folder path"):
        hook_with_fake.get_file_id("/")


def test_get_file_info_and_modified_time(hook_with_fake, box_fake):
    box_fake.seed_folder("/X")
    file_id = box_fake.seed_file(
        "/X/doc.pdf",
        size=123,
        modified_at="2025-01-01T00:00:00Z",
    )

    info = hook_with_fake.get_file_info("/X/doc.pdf")
    assert info.object_id == file_id
    assert info.name == "doc.pdf"

    file_id2, mod_time = hook_with_fake.get_file_modified_time("/X/doc.pdf")
    assert file_id2 == file_id
    assert mod_time == "2025-01-01T00:00:00Z"


def test_get_files_by_pattern_filters_and_converts(hook_with_fake, box_fake):
    box_fake.seed_folder("/data/sub")
    a_id = box_fake.seed_file("/data/a.txt")
    box_fake.seed_file("/data/b.csv")

    files = hook_with_fake.get_files_by_pattern("/data", "*.txt")
    assert [f.object_id for f in files] == [a_id]
    assert files[0].path.endswith("/a.txt")


def test_upload_file_new_file(tmp_path, hook_with_fake):
    local = tmp_path / "f.txt"
    local.write_text("hello")

    info = hook_with_fake.upload_file(str(local), "/dest/f.txt")
    assert info.object_id.isdigit()
    assert info.new is True
    assert hook_with_fake.get_file_id("/dest/f.txt") == info.object_id


def test_upload_file_updates_existing_file(tmp_path, hook_with_fake, box_fake):
    print("writing")
    local = tmp_path / "f.txt"
    local.write_text("hello")

    print("seed")
    existing_id = box_fake.seed_file("/dest/f.txt", content=b"old")

    print("uploading")
    info = hook_with_fake.upload_file(str(local), "/dest/f.txt")
    assert info.object_id == existing_id
    assert info.new is False

    downloaded = tmp_path / "out.txt"
    print("downloading")
    hook_with_fake.download_file(existing_id, str(downloaded))
    print("downloaded")
    assert downloaded.read_text() == "hello"


def test_upload_file_raises_when_folder_conflicts(tmp_path, hook_with_fake, box_fake):
    local = tmp_path / "f"
    local.write_text("x")

    box_fake.seed_folder("/dest/f")

    with pytest.raises(ValueError, match="folder with the name 'f' already exists"):
        hook_with_fake.upload_file(str(local), "/dest/f")


def test_create_folder_creates_remaining_path(hook_with_fake, box_fake):
    box_fake.seed_folder("/A")
    folder_id = hook_with_fake.create_folder("/A/B/C")
    assert hook_with_fake.get_folder_id("/A/B/C") == folder_id


def test_create_folder_noop_when_exists(hook_with_fake, box_fake):
    existing_id = box_fake.seed_folder("/A/B")

    folder_id = hook_with_fake.create_folder("/A/B")
    assert folder_id == existing_id


def test_download_file_by_id_writes_and_closes_stream(tmp_path, hook_with_fake, box_fake):
    file_id = box_fake.seed_file("/x.bin", content=b"abc")

    dest = tmp_path / "out.bin"
    info = hook_with_fake.download_file(file_id, str(dest))
    assert dest.read_bytes() == b"abc"
    assert info.object_id == file_id
    assert box_fake.last_download_stream is not None
    assert box_fake.last_download_stream.closed is True


def test_download_file_by_path_uses_get_file_id(tmp_path, hook_with_fake, box_fake):
    box_fake.seed_folder("/p")
    box_fake.seed_file("/p/f", content=b"x")

    dest = tmp_path / "out"
    hook_with_fake.download_file("/p/f", str(dest))
    assert dest.read_bytes() == b"x"


def test_delete_file_by_id_and_by_path(hook_with_fake, box_fake):
    file_id = box_fake.seed_file("/x/y")

    hook_with_fake.delete_file(file_id)
    with pytest.raises(KeyError):
        box_fake.get_file_by_id(file_id)

    file_id2 = box_fake.seed_file("/x/z")
    hook_with_fake.delete_file("/x/z")
    with pytest.raises(KeyError):
        box_fake.get_file_by_id(file_id2)


def test_box_file_to_file_info_path_collection_builds_path(box_fake):
    from box_airflow_provider.hooks.box import box_file_to_file_info

    file_id = box_fake.seed_file("/A/B/n.txt", size=0)
    box_file = box_fake.get_file_by_id(file_id)

    info = box_file_to_file_info(box_file)
    assert info.path == "/A/B/n.txt"
    assert info.size == 0



from dataclasses import dataclass
from datetime import datetime

@dataclass
class BoxTriggerEventData:
    """Model for BoxTrigger event data."""
    status: str
    message: str
    path: str
    files_sensed: list[tuple[str, str]] | None
    newer_than: datetime | str | None = None
    file_pattern: str = ""



from __future__ import annotations

import asyncio

import pytest

from tests.async_test_utils import patched_asyncio_for_tests


@pytest.mark.asyncio
async def test_patched_asyncio_for_tests_sleep_blocks_until_released_and_counts_calls():
    cm, controller = patched_asyncio_for_tests()

    with cm():
        sleep_task = asyncio.create_task(asyncio.sleep(0.01))

        await controller.wait_for_sleep_calls(1)
        assert controller.sleep_calls == 1
        assert sleep_task.done() is False

        controller.release_next_sleep()
        await sleep_task


@pytest.mark.asyncio
async def test_patched_asyncio_for_tests_to_thread_runs_inline():
    cm, _ = patched_asyncio_for_tests()

    with cm():
        result = await asyncio.to_thread(lambda x: x + 1, 41)
    assert result == 42



from __future__ import annotations

import sys
from pathlib import Path

from dataclasses import dataclass
from datetime import datetime
from types import SimpleNamespace
from typing import Any

import io

import pytest


@dataclass
class _FakeBoxItemType:
    value: str


@dataclass
class _FakePathCollection:
    entries: list[Any]


@dataclass
class FakeBoxItem:
    id: str
    name: str
    type: _FakeBoxItemType
    size: int | None = None
    created_at: datetime | None = None
    modified_at: datetime | None = None
    path_collection: _FakePathCollection | None = None


class FakeBoxEnvironment:
    """In-memory Box environment with deterministic IDs.

    This is intentionally a small surface-area fake tailored for unit tests.
    """

    ROOT_ID = "0"
    ROOT_NAME = "All Files"

    def __init__(self) -> None:
        self._next_id = 1
        """ Next available item ID """

        self._items: dict[str, FakeBoxItem] = {}
        """ Map of all item IDs to their metadata """

        self._children: dict[str, dict[str, str]] = {self.ROOT_ID: {}}
        """ Map of all parent IDs to their child IDs """

        self._file_content: dict[str, bytes] = {}
        """ Map of all file IDs to their content """

        self.last_download_stream: io.BytesIO | None = None
        """ Last downloaded file stream, if any """

        # Root folder exists implicitly.
        self._items[self.ROOT_ID] = FakeBoxItem(
            id=self.ROOT_ID,
            name=self.ROOT_NAME,
            type=_FakeBoxItemType("folder"),
            path_collection=_FakePathCollection(entries=[SimpleNamespace(name=self.ROOT_NAME)]),
        )

    def client(self) -> "FakeBoxClient":
        return FakeBoxClient(self)

    def seed_folder(self, path: str) -> str:
        return self._ensure_folder(path)

    def seed_file(
            self,
            path: str,
            *,
            content: bytes = b"",
            size: int | None = None,
            created_at: datetime | None = None,
            modified_at: datetime | None = None,
    ) -> str:
        folder_path, file_name = ("/" + path.strip("/")).rsplit("/", 1)
        folder_id = self._ensure_folder(folder_path)
        file_id = self._create_item(parent_id=folder_id, name=file_name, type_value="file")
        item = self._items[file_id]
        item.size = size if size is not None else len(content)
        item.created_at = created_at
        item.modified_at = modified_at
        item.path_collection = self._build_path_collection(parent_id=folder_id)
        self._file_content[file_id] = content
        return file_id

    def set_file_modified_at(self, file_id: str, modified_at: datetime | None) -> None:
        """Update a seeded file's modified time."""
        item = self._items[file_id]
        item.modified_at = modified_at

    def get_folder_items(self, folder_id: str, *, fields: list[str] | None = None) -> Any:
        children = self._children.get(folder_id, {})
        entries = [self._items[item_id] for item_id in children.values()]
        return SimpleNamespace(entries=entries)

    def get_file_by_id(self, file_id: str) -> FakeBoxItem:
        return self._items[file_id]

    def download_file(self, file_id: str) -> io.BytesIO:
        stream = io.BytesIO(self._file_content.get(file_id, b""))
        self.last_download_stream = stream
        return stream

    def delete_file_by_id(self, file_id: str) -> None:
        item = self._items.get(file_id)
        if item is None:
            return
        # Remove from parent children index
        for parent_id, children in self._children.items():
            for name, cid in list(children.items()):
                if cid == file_id:
                    del children[name]
                    break
        self._items.pop(file_id, None)
        self._file_content.pop(file_id, None)

    def create_folder(self, *, name: str, parent_id: str) -> FakeBoxItem:
        folder_id = self._create_item(parent_id=parent_id, name=name, type_value="folder")
        folder = self._items[folder_id]
        folder.path_collection = self._build_path_collection(parent_id=parent_id)
        return folder

    def upload_file(self, *, name: str, parent_id: str, file_obj: Any) -> Any:
        content = file_obj.read()
        file_id = self._create_item(parent_id=parent_id, name=name, type_value="file")
        item = self._items[file_id]
        item.size = len(content)
        item.path_collection = self._build_path_collection(parent_id=parent_id)
        self._file_content[file_id] = content
        return SimpleNamespace(entries=[item])

    def upload_file_version(self, *, file_id: str, name: str, file_obj: Any) -> Any:
        content = file_obj.read()
        item = self._items[file_id]
        item.name = name
        item.size = len(content)
        self._file_content[file_id] = content
        return SimpleNamespace(entries=[item])

    def _build_path_collection(self, *, parent_id: str) -> _FakePathCollection:
        # Build entries with at least a root element; hook code only reads `.name`.
        parts: list[str] = []
        current_id = parent_id
        while current_id != self.ROOT_ID:
            item = self._items[current_id]
            parts.append(item.name)
            current_id = getattr(item, "_parent_id")
        parts.reverse()
        entries = [SimpleNamespace(name=self.ROOT_NAME)] + [SimpleNamespace(name=p) for p in parts]
        return _FakePathCollection(entries=entries)

    def _ensure_folder(self, path: str) -> str:
        clean = path.strip("/")
        if not clean:
            return self.ROOT_ID
        current_id = self.ROOT_ID
        for part in clean.split("/"):
            if not part:
                continue
            children = self._children.setdefault(current_id, {})
            existing_id = children.get(part)
            if existing_id is not None:
                current_id = existing_id
                continue
            current_id = self._create_item(parent_id=current_id, name=part, type_value="folder")
            folder = self._items[current_id]
            folder.path_collection = self._build_path_collection(parent_id=getattr(folder, "_parent_id"))
        return current_id

    def _create_item(self, *, parent_id: str, name: str, type_value: str) -> str:
        new_id = str(self._next_id)
        self._next_id += 1
        item = FakeBoxItem(id=new_id, name=name, type=_FakeBoxItemType(type_value))
        # Keep parent pointer for path building.
        setattr(item, "_parent_id", parent_id)
        self._items[new_id] = item
        self._children.setdefault(parent_id, {})[name] = new_id
        if type_value == "folder":
            self._children.setdefault(new_id, {})
        return new_id


class _FakeUsersManager:
    """Fake for client.users"""

    def get_user_me(self) -> Any:  # pragma: no cover
        return SimpleNamespace(id="me")


class _FakeFoldersManager:
    """Fake for client.folders"""

    def __init__(self, env: FakeBoxEnvironment) -> None:
        self._env = env

    def get_folder_items(self, *, folder_id: str, fields: list[str] | None = None) -> Any:
        return self._env.get_folder_items(folder_id, fields=fields)

    def create_folder(self, *, name: str, parent: Any) -> FakeBoxItem:
        return self._env.create_folder(name=name, parent_id=parent.id)


class _FakeFilesManager:
    """Fake for client.files"""

    def __init__(self, env: FakeBoxEnvironment) -> None:
        self._env = env

    def get_file_by_id(self, *, file_id: str) -> FakeBoxItem:
        return self._env.get_file_by_id(file_id)

    def delete_file_by_id(self, *, file_id: str) -> None:
        self._env.delete_file_by_id(file_id)


class _FakeUploadsManager:
    """Fake for client.uploads"""

    def __init__(self, env: FakeBoxEnvironment) -> None:
        self._env = env

    def upload_file(self, *, attributes: Any, file: Any) -> Any:
        return self._env.upload_file(name=attributes.name, parent_id=attributes.parent.id, file_obj=file)

    def upload_file_version(self, *, file_id: str, attributes: Any, file: Any) -> Any:
        return self._env.upload_file_version(file_id=file_id, name=attributes.name, file_obj=file)


class _FakeDownloadsManager:
    """Fake for client.downloads"""

    def __init__(self, env: FakeBoxEnvironment) -> None:
        self._env = env

    def download_file(self, *, file_id: str) -> io.BytesIO:
        return self._env.download_file(file_id)


class FakeBoxClient:
    def __init__(self, env: FakeBoxEnvironment) -> None:
        self._env = env
        self.users = _FakeUsersManager()
        self.folders = _FakeFoldersManager(env)
        self.files = _FakeFilesManager(env)
        self.uploads = _FakeUploadsManager(env)
        self.downloads = _FakeDownloadsManager(env)


@pytest.fixture()
def box_fake() -> FakeBoxEnvironment:
    """Reusable fake Box environment with a known empty root (id='0')."""
    return FakeBoxEnvironment()


def pytest_configure() -> None:
    # Ensure `src/` is on `sys.path` so `box_airflow_provider` imports work when running pytest from repo root.
    project_root = Path(__file__).resolve().parents[1]
    src = project_root / "src"
    if str(src) not in sys.path:
        sys.path.insert(0, str(src))



import fnmatch
import logging
import os
from datetime import datetime
from os.path import relpath
from dataclasses import dataclass
from typing import Any, Literal
from collections.abc import Callable

from airflow.hooks.base import BaseHook
from box_sdk_gen import BoxClient, BoxCCGAuth, CCGConfig, FileFull
from box_sdk_gen.managers.uploads import UploadFileAttributes, UploadFileAttributesParentField, UploadFileVersionAttributes
from box_sdk_gen.managers.folders import CreateFolderParent

@dataclass
class BoxFileInfo:
    """
    Information about a Box file.

    The typing of the original Box SDK is far too complicated to be useful,
    so we're defining a simple model.
    """
    object_id: str
    name: str
    type: str
    size: int
    created_at: datetime | None
    modified_at: datetime | None
    path: str
    new: bool

class BoxHook(BaseHook):
    """
    Interact with Box API.

    This uses the official Box SDK for Python.

    :param box_conn_id: The connection ID to use for Box API
    """
    conn_name_attr = "box_conn_id"
    default_conn_name = "box_default"
    conn_type = "box"
    hook_name = "Box"

    @staticmethod
    def get_connection_form_widgets() -> dict[str, Any]:
        """Returns connection widgets to add to connection form"""
        from flask_appbuilder.fieldwidgets import BS3PasswordFieldWidget, BS3TextFieldWidget
        from flask_babel import lazy_gettext
        from wtforms import PasswordField, StringField

        return {
            "client_id": StringField(lazy_gettext("Client ID"), widget=BS3TextFieldWidget()),
            "client_secret": PasswordField(lazy_gettext("Client Secret"), widget=BS3PasswordFieldWidget()),
            "enterprise_id": PasswordField(lazy_gettext("Enterprise ID"), widget=BS3PasswordFieldWidget()),
        }

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        """Returns custom field behaviour"""

        return {
            "hidden_fields": ["port", "password", "login", "schema", "host"],
            "relabeling": {},
            "placeholders": {
                "client_id": "",
                "client_secret": "",
                "enterprise_id": ""
            },
        }

    def __init__(
            self,
            box_conn_id: str = default_conn_name,
            client: BoxClient | None = None,
            client_factory: Callable[[], BoxClient] | None = None,
    ) -> None:
        super().__init__()
        self.box_conn_id = box_conn_id
        self.client: BoxClient | None = client
        self.client_factory = client_factory

    def get_conn(self) -> BoxClient:
        if self.client_factory is not None:
            return self.client_factory()

        if self.box_conn_id and not self.client:
            conn = self.get_connection(self.box_conn_id)
            extra = conn.get_extra_dejson()
            client_id = extra['client_id']
            client_secret = extra['client_secret']
            enterprise_id = extra['enterprise_id']

            if not client_id or not client_secret or not enterprise_id:
                raise ValueError("Client ID, Client Secret and Enterprise ID must be provided in connection")

            ccg_config = CCGConfig(
                client_id=client_id,
                client_secret=client_secret,
                enterprise_id=enterprise_id,
            )
            auth = BoxCCGAuth(config=ccg_config)

            self.client = BoxClient(auth=auth)
            logging.getLogger("box_sdk_gen").setLevel(logging.WARNING)

        return self.client

    def test_connection(self) -> (bool, str):
        """Test a connection"""
        conn = self.get_conn()
        # Simple call to get_user_me to test the connection
        conn.users.get_user_me()
        return True, "Connection successfully tested"

    def get_item_id(self, path: str, item_type: Literal['file', 'folder']) -> str:
        """
        Get the ID of a Box item (file or folder) based on a path. Raises an exception if the item does not exist.

        :param path: Path to the item (e.g. "/Test/Area" or "/Test/Area/document.pdf")
        :param item_type: Type of item to look for, either 'file' or 'folder'
        :return: The ID of the target item if found
        """
        return self._get_item_id(path, item_type=item_type, return_deepest_existing=False)[0]

    def get_existing_item_id(self, path: str, item_type: Literal['file', 'folder']) -> tuple[str, str]:
        """
        Get the ID of an existing Box item (file or folder) based on a path.

        :param path: Path to the item (e.g. "/Test/Area" or "/Test/Area/document.pdf")
        :param item_type: Type of item to look for, either 'file' or 'folder'
        :return: The ID of the target item if found
        """
        return self._get_item_id(path, item_type=item_type, return_deepest_existing=True)

    def _get_item_id(self, path: str, item_type: Literal['file', 'folder'], return_deepest_existing: bool = False) -> \
            tuple[str, str]:
        """
        Get the ID of a Box item (file or folder) based on a path. Raises an exception if the item does not exist.

        If the item does not exist, it returns the ID of the deepest existing folder as well as the path to that folder.
        :return: A tuple with the ID and None, or the ID of the deepest existing folder and the path to that folder.
        """
        client = self.get_conn()

        current_folder_id = '0'
        current_path = '/'

        clean_path = path.strip('/')
        if not clean_path:
            if item_type == 'folder':
                return current_folder_id, None
            else:
                raise ValueError("Cannot get file ID for root folder path")

        path_parts = clean_path.split('/')

        # If looking for a file, the last part is the filename
        if item_type == 'file':
            folder_names = path_parts[:-1]
            file_name = path_parts[-1]
        else:
            folder_names = path_parts
            file_name = None

        for folder_name in folder_names:
            if not folder_name:
                continue

            items = client.folders.get_folder_items(folder_id=current_folder_id)

            found = False
            for item in items.entries:
                if item.type.value == 'folder' and item.name == folder_name:
                    current_folder_id = item.id
                    current_path = f"{current_path.rstrip('/')}/{folder_name}"
                    found = True
                    break

            if not found:
                if return_deepest_existing:
                    return current_folder_id, current_path
                raise FileNotFoundError(
                    f"Folder '{folder_name}' not found in path '{current_path}'"
                )

        # If looking for a file, find the file in the final folder
        if item_type == 'file' and file_name:
            items = client.folders.get_folder_items(folder_id=current_folder_id)

            for item in items.entries:
                if item.type.value == 'file' and item.name == file_name:
                    return item.id, None

            raise FileNotFoundError(f"File '{file_name}' not found in folder")

        if return_deepest_existing:
            return current_folder_id, current_path
        return current_folder_id, None

    def get_folder_id(self, path: str) -> str:
        """
        Get the ID of a Box folder based on a path.

        :param path: Path to the folder (e.g. "/Test/Area")
        :return: The ID of the target folder if found
        """
        return self.get_item_id(path, 'folder')

    def get_file_id(self, path: str) -> str:
        """
        Get the ID of a Box file based on a path.

        :param path: Path to the file (e.g. "/Test/Area/document.pdf")
        :return: The ID of the target file if found
        """
        return self.get_item_id(path, 'file')

    def get_file_info(self, path: str) -> BoxFileInfo:
        """
        Get the Box file information based on a path.

        :param path: Path to the file (e.g. "/Test/Area/document.pdf")
        :return: The file information if successful
        """
        file_id = self.get_file_id(path)

        client = self.get_conn()
        box_file = client.files.get_file_by_id(file_id=file_id)

        return box_file_to_file_info(box_file)

    def get_file_modified_time(self, path: str) -> tuple[str, str]:
        """
        Get the last modified time of a Box file based on a path.

        :param path: Path to the file (e.g. "/Test/Area/document.pdf")
        :return: The file's ID and last modified time if successful
        """
        file_info = self.get_file_info(path)

        return file_info.object_id, file_info.modified_at

    def get_files_by_pattern(self, path: str, file_pattern: str) -> list[BoxFileInfo]:
        """
        Get a list of Box files matching a pattern in a given folder.

        :param path: Path to the folder (e.g. "/Test/Area")
        :param file_pattern: Pattern to match files against (e.g. "*.pdf")
        :return: A list of matching BoxFileInfo objects
        """
        client = self.get_conn()
        folder_id = self.get_folder_id(path)

        items = client.folders.get_folder_items(folder_id=folder_id, fields=['name', 'type', 'size', 'created_at', 'modified_at', 'path_collection'])

        matching_files = []
        for item in items.entries:
            if item.type.value == 'file' and fnmatch.fnmatch(item.name, file_pattern):
                matching_files.append(box_file_to_file_info(item))

        return matching_files

    def upload_file(self, local_file_path: str, box_path: str) -> BoxFileInfo:
        """
        Upload a file to Box.

        :param local_file_path: Path to the local file to upload
        :param box_path: Destination path in Box (e.g. "/Test/Area/document.pdf")
        :return: The uploaded file information
        """
        client = self.get_conn()

        # Extract folder path and filename from box_path
        box_directory, filename = os.path.split(box_path.strip('/'))
        box_directory = '/' + box_directory if box_directory else '/'

        if not filename:
            raise ValueError("Must provide full path to Box file including filename")

        folder_id = self.create_folder(box_directory)

        # Check if file already exists and perform update if it does
        existing_items = client.folders.get_folder_items(folder_id=folder_id)
        file_already_exists = False
        existing_file_id = None

        for item in existing_items.entries:
            if item.name == filename:
                if item.type.value == 'file':
                    file_already_exists = True
                    existing_file_id = item.id
                elif item.type.value == 'folder':
                    raise ValueError(f"A folder with the name '{filename}' already exists in the destination path.")
                break

        if file_already_exists and existing_file_id:
            with open(local_file_path, 'rb') as file_stream:
                uploaded_file = client.uploads.upload_file_version(
                    file_id=existing_file_id, 
                    attributes=UploadFileVersionAttributes(name=filename), 
                    file=file_stream
                )
        else:
            with open(local_file_path, 'rb') as file_stream:
                uploaded_file = client.uploads.upload_file(
                    attributes=UploadFileAttributes(name=filename, parent=UploadFileAttributesParentField(id=folder_id)), 
                    file=file_stream
                )

        # Get the file info - the response is Files object with entries
        file_info = box_file_to_file_info(uploaded_file.entries[0])
        file_info.new = not file_already_exists

        return file_info

    def create_folder(self, path: str) -> str:
        """
        Create folders recursively based on the given path.

        :param path: The full path of the folder to create (e.g., "/Parent/Child/Grandchild").
        :return: The ID of the final folder created or found.
        """
        client = self.get_conn()

        # Find the deepest existing folder
        (id, expath) = self.get_existing_item_id(path, item_type='folder')

        current_folder_id = id
        current_path = expath or '/'

        # Get the remaining path to create
        remaining_path = relpath(path, current_path).strip('/')

        if remaining_path == '.':
            return current_folder_id

        # Create the remaining folders
        for folder_name in remaining_path.split('/'):
            folder = client.folders.create_folder(name=folder_name, parent=CreateFolderParent(id=current_folder_id))
            current_folder_id = folder.id

        return current_folder_id

    def download_file(self, box_path: str, local_file_path: str) -> BoxFileInfo:
        """
        Download a file from Box and save it to a local file path.

        :param box_path: Path to the file in Box (e.g., "/Test/Area/document.pdf") or a numeric string representing the file ID.
        :param local_file_path: Path to save the downloaded file locally.
        """
        # Check if box_path is a numeric string (file ID)
        if box_path.isdigit():
            file_id = box_path
        else:
            file_id = self.get_file_id(box_path)

        client = self.get_conn()
        box_file = client.files.get_file_by_id(file_id=file_id)
        info = box_file_to_file_info(box_file)

        file_content = client.downloads.download_file(file_id=file_id)
        # The download_file returns a file-like object, use context manager to ensure proper cleanup
        if file_content:
            try:
                with open(local_file_path, "wb") as local_file:
                    local_file.write(file_content.read())
            finally:
                # Ensure the stream is closed even if write fails
                if hasattr(file_content, 'close'):
                    file_content.close()

        return info

    def delete_file(self, box_path: str) -> None:
        """
        Delete a file from Box based on its path.

        :param box_path: Path to the file in Box (e.g., "/Test/Area/document.pdf") or a numeric string representing the file ID.
        """
        # Check if box_path is a numeric string (file ID)
        if box_path.isdigit():
            file_id = box_path
        else:
            file_id = self.get_file_id(box_path)

        client = self.get_conn()
        client.files.delete_file_by_id(file_id=file_id)


def box_file_to_file_info(box_file: FileFull) -> BoxFileInfo:
    """
    Convert a Box file object to a BoxFileInfo object.

    :param box_file: The Box file object to convert.
    :return: A BoxFileInfo object with the file's information.
    """
    # Build the path from path_collection if available
    if hasattr(box_file, 'path_collection') and box_file.path_collection and box_file.path_collection.entries:
        path_parts = [it.name for it in box_file.path_collection.entries[1::]]
        path = '/' + '/'.join(path_parts) + '/' + box_file.name
    else:
        path = '/' + box_file.name
    
    return BoxFileInfo(
        object_id=box_file.id,
        name=box_file.name,
        type=box_file.type.value,
        # Use the size attribute if available, otherwise default to 0
        size=box_file.size if hasattr(box_file, 'size') and box_file.size is not None else 0,
        created_at=box_file.created_at if hasattr(box_file, 'created_at') and box_file.created_at is not None else None,
        modified_at=box_file.modified_at if hasattr(box_file, 'modified_at') and box_file.modified_at is not None else None,
        path=path,
        new=False
    )



# read version from version.txt
from os import path

__version__ = open(path.join(path.dirname(__file__), '_version.txt')).read().strip()

## This is needed to allow Airflow to pick up specific metadata fields it needs for certain features.
def get_provider_info():
    return {
        "package-name": "apache-airflow-providers-box",
        "name": "Box",
        "description": "Apache Airflow provider for connecting to Box.com storage.",
        "connection-types": [
            {
                "connection-type": "box",
                "hook-class-name": "box_airflow_provider.hooks.box.BoxHook"
            }
        ],
        "versions": [__version__],
    }



from airflow.exceptions import AirflowException
from airflow.models import BaseOperator
from airflow.utils.context import Context

from box_airflow_provider.hooks.box import BoxHook, BoxFileInfo


class BoxUploadOperator(BaseOperator):
    template_fields = [
        "local_path",
        "box_path",
    ]

    def __init__(
            self,
            *,
            local_path: str | None = None,
            box_path: str | None = None,
            box_conn_id: str = BoxHook.default_conn_name,
            **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.box_conn_id = box_conn_id
        self.local_path = local_path
        self.box_path = box_path
        if not self.local_path:
            raise AirflowException("local_path is required")
        if not self.box_path:
            raise AirflowException("box_path is required")

    def execute(self, context: Context) -> BoxFileInfo:
        hook = BoxHook(box_conn_id=self.box_conn_id)

        response = hook.upload_file(self.local_path, self.box_path)

        return response


class BoxDownloadOperator(BoxUploadOperator):

    def execute(self, context: Context) -> BoxFileInfo:
        hook = BoxHook(box_conn_id=self.box_conn_id)

        response = hook.download_file(self.box_path, self.local_path)

        return response



from __future__ import annotations

from datetime import datetime, timezone
from types import SimpleNamespace
from unittest.mock import MagicMock, patch

import pendulum
import pytest
from box_airflow_provider.sensors.box import BoxSensor


def _file_info(*, path: str, name: str = "f", object_id: str = "1", modified_at=None):
    return SimpleNamespace(path=path, name=name, object_id=object_id, modified_at=modified_at)


def test_poke_with_pattern_returns_false_when_no_match():
    fake_hook = MagicMock()
    fake_hook.get_files_by_pattern.return_value = []
    sensor = BoxSensor(task_id="t", path="/data", file_pattern="*.csv", deferrable=False, hook=fake_hook)

    assert sensor.poke(context={}) is False


def test_poke_with_pattern_filters_newer_than_and_callable_returns_poke_return_value():
    newer_than = "2025-01-01T00:00:00Z"
    f_old = _file_info(path="/data/a.csv", object_id="a", modified_at="2024-12-31T23:59:59Z")
    f_new = _file_info(path="/data/b.csv", object_id="b", modified_at="2025-01-01T00:00:00Z")

    py_callable = MagicMock(return_value={"ok": True})
    fake_hook = MagicMock()
    fake_hook.get_files_by_pattern.return_value = [f_old, f_new]

    sensor = BoxSensor(
        task_id="t",
        path="/data",
        file_pattern="*.csv",
        newer_than=newer_than,
        hook=fake_hook,
        python_callable=py_callable,
        op_kwargs={"x": 1},
        deferrable=False,
    )

    result = sensor.poke(context={})

    assert result.is_done is True
    assert "files_found" in result.xcom_value
    assert [f.object_id for f in result.xcom_value["files_found"]] == ["b"]
    assert result.xcom_value["decorator_return_value"] == {"ok": True}
    py_callable.assert_called_once()


def test_poke_without_pattern_returns_true_when_file_exists():
    fake_hook = MagicMock()
    fake_hook.get_file_info.return_value = _file_info(
        path="/data/x.txt", modified_at=pendulum.datetime(2025, 1, 1, tz="UTC")
    )

    sensor = BoxSensor(task_id="t", path="/data/x.txt", deferrable=False, hook=fake_hook)
    assert sensor.poke(context={}) is True


def test_poke_invalid_newer_than_type_raises():
    fake_hook = MagicMock()
    fake_hook.get_file_info.return_value = _file_info(path="/x", modified_at=datetime.now(timezone.utc))

    sensor = BoxSensor(task_id="t", path="/x", newer_than=123, deferrable=False,
                       hook=fake_hook)  # type: ignore[arg-type]
    with pytest.raises(ValueError, match="Invalid newer_than value"):
        sensor.poke(context={})


def test_execute_non_deferrable_calls_base_execute():
    sensor = BoxSensor(task_id="t", path="/x", deferrable=False)
    with patch("airflow.sensors.base.BaseSensorOperator.execute") as base_execute:
        sensor.execute(context={})
    base_execute.assert_called_once()


def test_execute_defers_when_not_ready():
    sensor = BoxSensor(task_id="t", path="/x", deferrable=True, poke_interval=5)
    sensor.poke = MagicMock(return_value=False)
    sensor.defer = MagicMock()

    sensor.execute(context={})

    sensor.defer.assert_called_once()
    _, kwargs = sensor.defer.call_args
    assert kwargs["method_name"] == "execute_complete"
    trigger = kwargs["trigger"]
    assert trigger.path == "/x"
    assert trigger.poke_interval == 5


def test_execute_complete_success_returns_files_sensed():
    sensor = BoxSensor(task_id="t", path="/x", deferrable=True)
    event = {"status": "success", "message": "ok", "files_sensed": [("a", "1")]}
    assert sensor.execute_complete(context={}, event=event) == [("a", "1")]


def test_execute_complete_error_raises():
    from airflow.exceptions import AirflowException

    sensor = BoxSensor(task_id="t", path="/x", deferrable=True)
    event = {"status": "error", "message": "bad", "files_sensed": None}
    with pytest.raises(AirflowException, match="bad"):
        sensor.execute_complete(context={}, event=event)



from __future__ import annotations

from contextlib import ExitStack, contextmanager, _GeneratorContextManager
from dataclasses import dataclass, field
from typing import Any, Callable
import asyncio
from unittest.mock import patch


@dataclass
class SleepController:
    """Controller used by tests to coordinate patched `asyncio.sleep`.

    The patched sleep will block until the test releases it.
    """

    sleep_calls: int = 0
    _waiters: list[asyncio.Future[None]] = field(default_factory=list)
    _calls_changed: asyncio.Event = field(default_factory=asyncio.Event)

    async def wait_for_sleep_calls(self, at_least: int = 1) -> None:
        while self.sleep_calls < at_least:
            self._calls_changed.clear()
            await self._calls_changed.wait()

    def release_next_sleep(self) -> None:
        if not self._waiters:
            return
        fut = self._waiters.pop(0)
        if not fut.done():
            fut.set_result(None)

    def release_all_sleeps(self) -> None:
        while self._waiters:
            self.release_next_sleep()


def patched_asyncio_for_tests(
    *,
    sleep_patch_target: str = "asyncio.sleep",
    to_thread_patch_target: str = "asyncio.to_thread",
) -> tuple[Callable[..., _GeneratorContextManager[Any, None, None]], SleepController]:
    """
    Return a contextmanager that patches `asyncio.sleep`/`asyncio.to_thread`
    and a sleep counter.

    The returned context manager can be used in a `with` block. While active:
    - `asyncio.sleep` is replaced with an awaitable that blocks until the test releases it via
      the returned `SleepController`.
    - `asyncio.to_thread` is replaced with an awaitable that calls the function synchronously.

    `sleep_patch_target` / `to_thread_patch_target` allow patching the module-under-test, e.g.
    `"box_airflow_provider.triggers.box.asyncio.sleep"`.
    """

    controller = SleepController()

    async def _patched_sleep(delay: float | int = 0, result: Any = None) -> Any:  # noqa: ARG001
        controller.sleep_calls += 1
        controller._calls_changed.set()

        loop = asyncio.get_running_loop()
        waiter: asyncio.Future[None] = loop.create_future()
        controller._waiters.append(waiter)
        await waiter
        return result

    async def _patched_to_thread(func: Callable[..., Any], /, *args: Any, **kwargs: Any) -> Any:
        return func(*args, **kwargs)

    @contextmanager
    def _cm():
        with ExitStack() as stack:
            stack.enter_context(patch(sleep_patch_target, new=_patched_sleep))
            stack.enter_context(patch(to_thread_patch_target, new=_patched_to_thread))
            yield

    return _cm, controller




from __future__ import annotations

import asyncio
from asyncio import Task
from datetime import datetime, timezone

import pendulum
import pytest

from box_airflow_provider.hooks.box import BoxHook
from box_airflow_provider.triggers.box import BoxTrigger

from tests.async_test_utils import patched_asyncio_for_tests


def _hook_for(box_fake) -> BoxHook:
    return BoxHook(client_factory=box_fake.client)


async def _assert_trigger_sleeps(trigger: BoxTrigger, controller) -> None:
    gen = trigger.run()
    next_event_task: Task = asyncio.create_task(gen.__anext__())
    await controller.wait_for_sleep_calls(1)
    assert next_event_task.done() is False
    next_event_task.cancel()
    controller.release_all_sleeps()
    with pytest.raises(asyncio.CancelledError):
        await next_event_task


@pytest.mark.asyncio
async def test_run_with_pattern_no_files_continues_to_sleep(box_fake):
    box_fake.seed_folder("/data")

    trigger = BoxTrigger(path="/data", file_pattern="*.csv", poke_interval=0.01, hook=_hook_for(box_fake))

    cm, controller = patched_asyncio_for_tests(
        sleep_patch_target="box_airflow_provider.triggers.box.asyncio.sleep",
        to_thread_patch_target="box_airflow_provider.triggers.box.asyncio.to_thread",
    )
    with cm():
        await _assert_trigger_sleeps(trigger, controller)


@pytest.mark.asyncio
async def test_run_with_pattern_files_dont_match_continues_to_sleep(box_fake):
    box_fake.seed_file("/data/x.txt", modified_at=pendulum.datetime(2025, 1, 1, tz="UTC"))

    trigger = BoxTrigger(path="/data", file_pattern="*.csv", poke_interval=0.01, hook=_hook_for(box_fake))

    cm, controller = patched_asyncio_for_tests(
        sleep_patch_target="box_airflow_provider.triggers.box.asyncio.sleep",
        to_thread_patch_target="box_airflow_provider.triggers.box.asyncio.to_thread",
    )
    with cm():
        await _assert_trigger_sleeps(trigger, controller)


@pytest.mark.asyncio
async def test_run_with_pattern_matching_files_but_older_continues_to_sleep(box_fake):
    newer_than = pendulum.datetime(2025, 1, 1, tz="UTC")
    box_fake.seed_file("/data/f.csv", modified_at=pendulum.datetime(2024, 12, 31, 23, 59, 59, tz="UTC"))

    trigger = BoxTrigger(
        path="/data",
        file_pattern="*.csv",
        newer_than=newer_than,
        poke_interval=0.01,
        hook=_hook_for(box_fake),
    )

    cm, controller = patched_asyncio_for_tests(
        sleep_patch_target="box_airflow_provider.triggers.box.asyncio.sleep",
        to_thread_patch_target="box_airflow_provider.triggers.box.asyncio.to_thread",
    )
    with cm():
        await _assert_trigger_sleeps(trigger, controller)


@pytest.mark.asyncio
async def test_run_with_pattern_matching_files_and_newer_triggers(box_fake):
    newer_than = pendulum.datetime(2025, 1, 1, tz="UTC")
    file_id = box_fake.seed_file("/data/f.csv", modified_at=pendulum.datetime(2025, 1, 1, 0, 0, 0, tz="UTC"))

    trigger = BoxTrigger(
        path="/data",
        file_pattern="*.csv",
        newer_than=newer_than,
        poke_interval=0.01,
        hook=_hook_for(box_fake),
    )

    cm, controller = patched_asyncio_for_tests(
        sleep_patch_target="box_airflow_provider.triggers.box.asyncio.sleep",
        to_thread_patch_target="box_airflow_provider.triggers.box.asyncio.to_thread",
    )
    with cm():
        gen = trigger.run()
        event = await asyncio.wait_for(gen.__anext__(), timeout=1)

    assert controller.sleep_calls == 0
    assert event.payload["status"] == "success"
    assert event.payload["files_sensed"] == [("f.csv", file_id)]


@pytest.mark.asyncio
async def test_run_folder_does_not_exist_triggers_error(box_fake):
    trigger = BoxTrigger(path="/missing", file_pattern="*.csv", poke_interval=0.01, hook=_hook_for(box_fake))

    cm, controller = patched_asyncio_for_tests(
        sleep_patch_target="box_airflow_provider.triggers.box.asyncio.sleep",
        to_thread_patch_target="box_airflow_provider.triggers.box.asyncio.to_thread",
    )
    with cm():
        gen = trigger.run()
        event = await asyncio.wait_for(gen.__anext__(), timeout=1)

    assert controller.sleep_calls == 0
    assert event.payload["status"] == "error"
    assert "FileNotFoundError" in event.payload["message"]


@pytest.mark.asyncio
async def test_run_initially_older_then_updated_triggers(box_fake):
    newer_than = datetime(2025, 1, 1, tzinfo=timezone.utc)
    file_id = box_fake.seed_file(
        "/data/f.csv",
        modified_at=datetime(2024, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
    )

    trigger = BoxTrigger(
        path="/data",
        file_pattern="*.csv",
        newer_than=newer_than,
        poke_interval=0.01,
        hook=_hook_for(box_fake),
    )

    cm, controller = patched_asyncio_for_tests(
        sleep_patch_target="box_airflow_provider.triggers.box.asyncio.sleep",
        to_thread_patch_target="box_airflow_provider.triggers.box.asyncio.to_thread",
    )
    with cm():
        gen = trigger.run()
        next_event_task = asyncio.create_task(gen.__anext__())

        await controller.wait_for_sleep_calls(1)
        assert next_event_task.done() is False

        box_fake.set_file_modified_at(file_id, datetime(2025, 1, 1, 0, 0, 0, tzinfo=timezone.utc))

        controller.release_next_sleep()
        event = await asyncio.wait_for(next_event_task, timeout=1)

    assert event.payload["status"] == "success"
    assert event.payload["files_sensed"] == [("f.csv", file_id)]


@pytest.mark.asyncio
async def test_run_without_pattern_file_not_found_continues_to_sleep(box_fake):
    trigger = BoxTrigger(path="/data/x.txt", file_pattern="", poke_interval=0.01, hook=_hook_for(box_fake))

    cm, controller = patched_asyncio_for_tests(
        sleep_patch_target="box_airflow_provider.triggers.box.asyncio.sleep",
        to_thread_patch_target="box_airflow_provider.triggers.box.asyncio.to_thread",
    )
    with cm():
        await _assert_trigger_sleeps(trigger, controller)


@pytest.mark.asyncio
async def test_run_without_pattern_file_exists_and_newer_triggers(box_fake):
    file_id = box_fake.seed_file("/data/x.txt", modified_at=pendulum.datetime(2025, 1, 1, tz="UTC"))

    trigger = BoxTrigger(
        path="/data/x.txt",
        newer_than="2025-01-01T00:00:00Z",
        poke_interval=0.01,
        hook=_hook_for(box_fake),
    )

    cm, controller = patched_asyncio_for_tests(
        sleep_patch_target="box_airflow_provider.triggers.box.asyncio.sleep",
        to_thread_patch_target="box_airflow_provider.triggers.box.asyncio.to_thread",
    )
    with cm():
        gen = trigger.run()
        event = await asyncio.wait_for(gen.__anext__(), timeout=1)

    assert controller.sleep_calls == 0
    assert event.payload["status"] == "success"
    assert event.payload["files_sensed"] == [("x.txt", file_id)]



from datetime import datetime, timedelta
from typing import Callable, Any

import pendulum
from airflow.configuration import conf
from airflow.exceptions import AirflowException
from airflow.sensors.base import BaseSensorOperator, PokeReturnValue
from airflow.utils.context import Context
from airflow.utils.timezone import parse, convert_to_utc

from box_airflow_provider.hooks.box import BoxHook
from box_airflow_provider.triggers.box import BoxTrigger


class BoxSensor(BaseSensorOperator):
    """
    Waits for a file or folder to be available in Box.

    :param path: Box file or folder path
    :param file_pattern: The pattern that will be used to match the file (fnmatch format)
    :param box_conn_id: The connection to run the sensor against
    :param newer_than: DateTime for which the file or file path should be newer than, comparison is inclusive
    :param deferrable: If waiting for completion, whether to defer the task until done, default is ``False``.
    """

    template_fields = [
        "path",
        "newer_than",
    ]

    def __init__(
            self,
            *,
            box_conn_id: str = BoxHook.default_conn_name,
            path: str,
            file_pattern: str = "",
            newer_than: datetime | str | None = None,
            hook: BoxHook | None = None,
            python_callable: Callable | None = None,
            op_args: list | None = None,
            op_kwargs: dict[str, any] | None = None,
            deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
            **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.path = path
        self.file_pattern = file_pattern
        self.hook: BoxHook | None = hook
        self.box_conn_id = box_conn_id
        self.newer_than: datetime | str | None = newer_than
        self.python_callable: Callable | None = python_callable
        self.op_args = op_args or []
        self.op_kwargs = op_kwargs or {}
        self.deferrable = deferrable

    def poke(self, context: Context) -> PokeReturnValue | bool:
        if self.hook is None:
            self.hook = BoxHook(self.box_conn_id)
        self.log.debug("Poking for %s, with pattern %s", self.path, self.file_pattern)
        files_found = []

        if self.file_pattern:
            files_from_pattern = self.hook.get_files_by_pattern(self.path, self.file_pattern)
            if files_from_pattern:
                actual_files_to_check = files_from_pattern
            else:
                return False
        else:
            actual_files_to_check = [self.hook.get_file_info(self.path)]

        for file_info in actual_files_to_check:
            if not file_info:
                continue

            mod_time = file_info.modified_at

            if self.newer_than and self.newer_than != "None":
                if isinstance(self.newer_than, str):
                    _newer_than = parse(self.newer_than)
                elif isinstance(self.newer_than, datetime):
                    _newer_than = pendulum.instance(self.newer_than)
                else:
                    raise ValueError(f"Invalid newer_than value: {self.newer_than}")

                if isinstance(mod_time, str):
                    _mod_time = parse(mod_time)
                else:
                    _mod_time = pendulum.instance(mod_time)


                if _newer_than <= _mod_time:
                    files_found.append(file_info)
                    self.log.info(
                        "File %s has modification time: '%s', which is newer than: '%s'",
                        file_info.path,
                        str(_mod_time),
                        str(_newer_than),
                    )
                else:
                    self.log.info(
                        "File %s has modification time: '%s', which is older than: '%s'",
                        file_info.path,
                        str(_mod_time),
                        str(_newer_than),
                    )
            else:
                files_found.append(file_info)

        if not len(files_found):
            return False

        if self.python_callable is not None:
            if self.op_kwargs:
                self.op_kwargs["files_found"] = files_found
            callable_return = self.python_callable(*self.op_args, **self.op_kwargs)
            return PokeReturnValue(
                is_done=True,
                xcom_value={"files_found": files_found, "decorator_return_value": callable_return},
            )
        return True

    def execute(self, context: Context) -> None:
        if not self.deferrable:
            super().execute(context=context)
        elif not self.poke(context):
            self.defer(
                timeout=timedelta(seconds=self.timeout),
                trigger=BoxTrigger(
                    path=self.path,
                    file_pattern=self.file_pattern,
                    box_conn_id=self.box_conn_id,
                    poke_interval=self.poke_interval,
                    newer_than=self.newer_than,
                ),
                method_name="execute_complete",
            )

    def execute_complete(self, context: Context, event: dict[str, Any] | None = None) -> None:
        if event["status"] == "success":
            self.log.info(event["message"])
            return event["files_sensed"]
        else:
            raise AirflowException(event["message"])



# ASF Adapted from https://github.com/apache/airflow/blob/providers-sftp/4.11.1/airflow/providers/sftp/triggers/sftp.py

import asyncio
import traceback
from dataclasses import asdict
from datetime import datetime
from typing import Any, AsyncIterator

import pendulum
from airflow.triggers.base import BaseTrigger, TriggerEvent
from airflow.utils.timezone import parse

from box_airflow_provider.hooks.box import BoxHook, BoxFileInfo
from box_airflow_provider.models import BoxTriggerEventData


class BoxTrigger(BaseTrigger):
    def __init__(
            self,
            path: str,
            file_pattern: str = "",
            box_conn_id: str = BoxHook.default_conn_name,
            newer_than: datetime | str | None = None,
            poke_interval: float = 60,
            hook: BoxHook | None = None,
            files_sensed: list[tuple[str, str]] | None = None,
            status: str = "",
            message: str = "",
    ) -> None:
        super().__init__()
        self.path = path
        self.file_pattern = file_pattern
        self.box_conn_id = box_conn_id
        self.newer_than = newer_than
        self.poke_interval = poke_interval
        self.hook = hook
        self.files_sensed = files_sensed
        self.status = status
        self.message = message

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize BoxTrigger arguments and classpath."""
        return (
            "box_airflow_provider.triggers.box.BoxTrigger",
            {
                "path": self.path,
                "file_pattern": self.file_pattern,
                "box_conn_id": self.box_conn_id,
                "newer_than": self.newer_than,
                "poke_interval": self.poke_interval,
            },
        )

    def _create_event_data(self, status: str, message: str, files_sensed: list[tuple[str, str]] | None = None) -> BoxTriggerEventData:
        """
        Utility function to create a BoxTriggerEventData object using instance variables.

        :param status: Status of the event (e.g., "success", "error").
        :param message: Message describing the event.
        :param files_sensed: List of files sensed, if any.
        :return: A BoxTriggerEventData object.
        """
        return BoxTriggerEventData(
            status=status,
            message=message,
            newer_than=self.newer_than,
            files_sensed=files_sensed,
            path=self.path,
            file_pattern=self.file_pattern,
        )

    def yield_even_data(self, status: str, message: str, files_sensed: list[tuple[str, str]] | None = None) -> TriggerEvent:
        return TriggerEvent(asdict(self._create_event_data(status, message, files_sensed)))

    async def run(self) -> AsyncIterator[TriggerEvent]:
        """
        Make a series of asynchronous calls to Box API. It yields a TriggerEvent.

        - If file matching file pattern exists at the specified path, return it.
        - If file pattern was not provided, it looks directly into the specific path provided.
        - If newer_than datetime was provided, it checks the file's last modified time.
        """
        hook = self.hook or BoxHook(self.box_conn_id)
        if isinstance(self.newer_than, str):
            self.newer_than = parse(self.newer_than)
        _newer_than = pendulum.instance(self.newer_than) if self.newer_than else None

        while True:
            try:
                if self.file_pattern:
                    files_result: list[BoxFileInfo] = await asyncio.to_thread(
                        hook.get_files_by_pattern, self.path, self.file_pattern
                    )
                    if not files_result or len(files_result) == 0:
                        await asyncio.sleep(self.poke_interval)
                        continue

                    files_sensed = []
                    for file_info in files_result:
                        if _newer_than:
                            mod_time = pendulum.instance(file_info.modified_at)
                            if _newer_than <= mod_time:
                                files_sensed.append((file_info.name, file_info.object_id))
                        else:
                            files_sensed.append((file_info.name, file_info.object_id))

                    if files_sensed:
                        event_data = self._create_event_data(
                            status="success",
                            message=f"Sensed {len(files_sensed)} files: {files_sensed}",
                            files_sensed=files_sensed,
                        )
                        yield self.yield_even_data(event_data.status, event_data.message, event_data.files_sensed)
                        return
                else:
                    # Catch FileNotFoundError and sleep, other errors should propagate
                    try:
                        file_info_result: BoxFileInfo = await asyncio.to_thread(
                            hook.get_file_info, self.path
                        )
                    except FileNotFoundError:
                        await asyncio.sleep(self.poke_interval)
                        continue

                    if not file_info_result:
                        await asyncio.sleep(self.poke_interval)
                        continue

                    file_info = file_info_result
                    if _newer_than:
                        mod_time = pendulum.instance(file_info.modified_at)
                        if _newer_than <= mod_time:
                            event_data = self._create_event_data(
                                status="success",
                                message=f"Sensed file: {self.path}",
                                files_sensed=[(file_info.name, file_info.object_id)],
                            )
                            yield self.yield_even_data(event_data.status, event_data.message, event_data.files_sensed)
                            return
                    else:
                        event_data = self._create_event_data(
                            status="success",
                            message=f"Sensed file: {self.path}",
                            files_sensed=[(file_info.name, file_info.object_id)],
                        )
                        yield self.yield_even_data(event_data.status, event_data.message, event_data.files_sensed)
                        return

                await asyncio.sleep(self.poke_interval)
            except Exception as e:
                event_data = self._create_event_data(
                    status="error",
                    message=traceback.format_exc(),
                    files_sensed=None,
                )
                yield self.yield_even_data(event_data.status, event_data.message, event_data.files_sensed)
                return
